#include <iostream>
#include <iomanip>
#include <stdlib.h>
#include <algorithm>
#include <time.h>
#include "SchedulingAlgorithm.h"

using namespace std;

const int    PRIORITY_MIN = 1;
const int    PRIORITY_MAX = 4;

const double RUNTIME_MIN  =  0.1;
const double RUNTIME_MAX  = 10.0;

void SchedulingAlgorithm::run_simulations()
{
    cout << endl << "=============================="
         << endl << "==============================" << endl;

    for (int i = 1; i <= run_count; i++)
    {
        ready_queue.clear();
        completed.clear();
        timeline = "";

        prepare_run(i);
        initialize_job_queue();
        print_job_queue();
        run_algorithm();
        print_timeline();
        print_statistics(i);
    }

    cout << endl;
    cout << "--------------------" << endl;
    cout << "Averages over " << run_count << " runs" << endl;
    cout << "--------------------" << endl;
    printf("Average wait time  = %6.2f\n", stats.average_wait_time);
    printf("Average turnaround = %6.2f\n", stats.average_turnaround);
    printf("Average throughput = %6.2f\n", stats.average_throughput);
}

void SchedulingAlgorithm::initialize_job_queue()
{
    job_queue.clear();

    for (int i = 0; i < process_count; i++)
    {
        int priority = rand()%PRIORITY_MAX + PRIORITY_MIN;
        double arrival_time = quantum_limit*(static_cast<double>(rand())/RAND_MAX);
        double run_time =
            (RUNTIME_MAX - RUNTIME_MIN)*(static_cast<double>(rand())/RAND_MAX)
            + RUNTIME_MIN;

        job_queue.push_back(new Process(priority, arrival_time, run_time));
    }

    // Sort by arrival time.
    sort(job_queue.begin(), job_queue.end(),
         [] (Process *p1, Process *p2) { return p1->arrival_time < p2->arrival_time; });

    start_time = job_queue[0]->arrival_time;

    // Assign names in sorted order.
    char ch = 'A';
    for (Process *p : job_queue) p->name = ch++;
}

void SchedulingAlgorithm::print_job_queue()
{
    cout << endl;
    cout << "Job queue:" << endl << endl;
    cout << "    Name  Pri  Arrive  Exp.RT" << endl;

    int i = 1;
    for (Process *p : job_queue)
    {
        printf("%3d    %c    %d   %5.2f    %4.2f\n",
                i++, p->name, p->priority, p->arrival_time,
                p->expected_run_time);
    }
}

void SchedulingAlgorithm::print_timeline()
{
    cout << endl << "Timeline:" << endl;
    cout << " ";
    for (int i = 1; i <= 12; i++) cout << setw(10) << i;
    cout << endl;
    for (int i = 0; i <= 11; i++) cout << "0123456789";
    cout << "0" << endl;
    for (int i = 0; i < timeline.size(); i++) cout << timeline[i];
    cout << endl;
}

void SchedulingAlgorithm::print_statistics(int run_index)
{
    double total_wait       = 0.0;
    double total_turnaround = 0.0;

    cout << endl;
    cout << "Process statistics:" << endl << endl;
    cout << "    Wait  Turnaround" << endl;

    // Sort by process name.
    sort(completed.begin(), completed.end(),
         [] (Process *p1, Process *p2) { return p1->name < p2->name; });

    for (Process *p : completed)
    {
        printf("%c  %5.2f  %10.2f\n",
                p->name, p->stats.wait_time, p->stats.turnaround_time);

        total_wait += p->stats.wait_time;
        total_turnaround += p->stats.turnaround_time;
    }

    double avg_wait_time  = total_wait/completed.size();
    double avg_turnaround = total_turnaround/completed.size();
    double elapsed_time = end_time - start_time;
    double throughput = (quantum_limit + 1)*completed.size()/elapsed_time;
    int not_run = process_count - completed.size();

    cout << endl;
    printf("Average wait time  = %5.2f\n", avg_wait_time);
    printf("Average turnaround = %5.2f\n", avg_turnaround);
    printf("Processes run      = %ld\n", completed.size());
    printf("Processes not run  = %d\n", not_run);

    cout << endl;
    printf("  Start time = %6.2f\n", start_time);
    printf("    End time = %6.2f\n", end_time);
    printf("Elapsed time = %6.2f\n", elapsed_time);
    printf("  Throughput = %6.2f\n", throughput);

    stats.average_wait_time =
        ((run_index - 1)*stats.average_wait_time + avg_wait_time)/run_index;
    stats.average_turnaround =
        ((run_index - 1)*stats.average_turnaround + avg_turnaround)/run_index;
    stats.average_throughput =
        ((run_index - 1)*stats.average_throughput + throughput)/run_index;
}
